/*
 * URBSession.java
 * Created on 19-Sep-2003, 11:37:01
 */
package protocols.uniformReliableBroadcast;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;

import protocols.utils.MessageEntry;
import protocols.utils.MessageID;

import utils.Debug;
import utils.ProcessSet;
import utils.SampleProcess;
import events.Crash;
import events.ProcessInitEvent;
import appia.AppiaEventException;
import appia.Direction;
import appia.Event;
import appia.Layer;
import appia.Session;
import appia.events.SendableEvent;
import appia.events.channel.ChannelInit;
import appia.message.ExtendedMessage;
import appia.protocols.common.InetWithPort;

/**
 * Class URBSession
 * @author nuno
 */
public class URBSession extends Session {

	// List of processes.
	private ProcessSet processes;
	// This sequence number represents the delivered set.
	private int seqNumber;
	// List of MessageID objects
	private LinkedList received, delivered;
	// hashtable mapping: < MessageID --> MessageEntry >
	private Hashtable ack;
	

	/**
	 * @param layer
	 */
	public URBSession(Layer layer) {
		super(layer);
	}

	public void handle(Event event) {
		// Init events. Channel Init is from Appia and ProcessInitEvent is to know the elements of the group
		if (event instanceof ChannelInit)
			handleChannelInit((ChannelInit) event);
		else if (event instanceof ProcessInitEvent)
			handleProcessInitEvent((ProcessInitEvent) event);
		else if (event instanceof SendableEvent) {
			if (event.getDir() == Direction.DOWN)
				// UPON event from the above protocol (or application)
				urbBroadcast((SendableEvent) event);
			else
				// UPON event from the bottom protocol (or perfect point2point links)
				bebDeliver((SendableEvent) event);
		} else if (event instanceof Crash)
			handleCrash((Crash) event);
		/*
		 * Unexpected event arrived. Forwarding it.
		 */
		else
			try {
				event.go();
			} catch (AppiaEventException e) {
				e.printStackTrace();
			}		

		// Every time something happens, the protocol verify if more messages can be delivered.
		urbTryDeliver();
	}

	private void urbTryDeliver() {
		Iterator it = ack.values().iterator();
		MessageEntry entry=null;
		while( it.hasNext() ){
			entry = (MessageEntry) it.next();
			if(canDeliver(entry)){
				delivered.add(entry.messageID);
				urbDeliver(entry.event, entry.messageID.process);
			}
		}
	}

	/**
	 * @param entry
	 * @return
	 */
	private boolean canDeliver(MessageEntry entry) {
		int procSize = processes.getSize();
		for(int i=0; i<procSize; i++)
			if(processes.getProcess(i).isCorrect() && (! entry.acks[i]) )
				return false;
		return ( ! delivered.contains(entry.messageID) );
	}

	/**
	 * @param init
	 */
	private void handleChannelInit(ChannelInit init) {
		try {
			init.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
		received = new LinkedList();
		delivered = new LinkedList();
		ack = new Hashtable();
	}

	/**
	 * @param event
	 */
	private void handleProcessInitEvent(ProcessInitEvent event) {
		processes = event.getProcessSet();
		try {
			event.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Called when the above protocol sends a message.
	 * @param event
	 */
	private void urbBroadcast(SendableEvent event) {
		/* header is composed by a sequence number
		 * and the number of the process that is sending the message
		 */
		SampleProcess self = processes.getSelfProcess();
		MessageID msgID = new MessageID(self.getProcessNumber(),seqNumber);
		Debug.print("URB: broadcasting message from "+ msgID.process +"with seqNumber = "+msgID.seqNumber);
		seqNumber++; 
		received.add(msgID);
		((ExtendedMessage) event.getMessage()).pushObject(msgID);
		try {
			event.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Called when the lower protocol delivers a message.
	 * @param event
	 */
	private void bebDeliver(SendableEvent event) {
		Debug.print("URB: Received message from beb.");
		SendableEvent clone = null;
		try {
			clone = (SendableEvent) event.cloneEvent();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
			return;
		}
		MessageID msgID = (MessageID) ((ExtendedMessage) clone.getMessage()).popObject();
		addAck(clone,msgID);
		if( ! received.contains(msgID) ){
			Debug.print("URB: Message is not on the received set.");
			received.add(msgID);
			bebBroadcast(event);
		}
	}

	/**
	 * Called by this protocol to send a message to the lower protocol.
	 * @param event
	 */
	private void bebBroadcast(SendableEvent event) { 
		Debug.print("URB: sending message to beb.");
		try {
			event.setDir(Direction.DOWN);
			event.setSource(this);
			event.init();
			event.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Delivers the message to above protocol or application
	 * @param event
	 * @param self
	 */
	private void urbDeliver(SendableEvent event, int sender) {
		Debug.print("URB: delivering message to above protocol.");
		try {
			event.setDir(Direction.UP);
			event.setSource(this);
			event.source = processes.getProcess(sender).getInetWithPort();
			event.init();
			event.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Called when some process crashed.
	 * @param crash
	 */
	private void handleCrash(Crash crash) {
		int crashedProcess = crash.getCrashedProcess();
		System.out.println("Process "+crashedProcess+" failed.");
		// changes the state of the process to "failed"
		processes.getProcess(crashedProcess).setCorrect(false);
	}

	private void addAck(SendableEvent event, MessageID msgID){
		Debug.print("URB: adding ack.");
		int pi = processes.getProcess((InetWithPort)event.source).getProcessNumber();
		MessageEntry entry = (MessageEntry) ack.get(msgID);
		if(entry == null){
			Debug.print("URB: first time that the message is seen.");
			entry = new MessageEntry(event, msgID, processes.getSize());
			ack.put(msgID,entry);
		}
		entry.acks[pi] = true;
	}
}


